#!/bin/env python

import collections.abc
import os
import hashlib
import numpy as np
import h5py
import unyt

import lightcone_io.units
from lightcone_io.utils import LocalOrRemoteFile


class HealpixMap(collections.abc.Sequence, LocalOrRemoteFile):
    """
    Class used to read lightcone HEALPix maps written out by SWIFT.
    Numpy-style indexing can be used to read pixels from the map.

    :param filenames: names of the HDF5 files containing map data
    :type  filenames: list of str
    :param map_name: name of the map to read
    :type  map_name: str
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, filenames, map_name, remote_dir=None):
        self.set_directory(remote_dir)
        self._filenames = filenames
        self._map_name  = map_name
        self._nr_pixels = None

    def _set_metadata(self):
        if self._nr_pixels is None:
            with self.open_file(self._filenames[0]) as infile:
                self._nr_pixels = infile[self._map_name].attrs["number_of_pixels"][0]
                self._nside = infile[self._map_name].attrs["nside"][0]
                self._dtype = infile[self._map_name].dtype
                self._pix_per_file = infile[self._map_name].shape[0]
                if unyt is not None:
                    self._units = lightcone_io.units.units_from_attributes(infile[self._map_name])

    @property
    def nside(self):
        """
        Return the nside (resolution) parameter of this HEALPix map
        """
        self._set_metadata()
        return self._nside

    @property
    def dtype(self):
        """
        Return the numpy data type of this HEALPix map
        """
        self._set_metadata()
        return self._dtype

    @property
    def units(self):
        """
        Return the unyt units of this HEALPix map
        """
        self._set_metadata()
        return self._units

    def __len__(self):
        self._set_metadata()
        return self._nr_pixels

    def read_pixels(self, start, stop):
        """
        Read the specified range of pixels from the map and return a numpy
        array.

        :param start: index of the first pixel to read
        :type  start: int
        :param stop: index just after the last pixel to read
        :type  stop: int

        :return: a numpy array with the specified pixel data
        :rtype: numpy.ndarray
        """
        self._set_metadata()

        if stop < start:
            raise KeyError("range of pixels to read must have stop >= start")

        if start < 0 or start >= self._nr_pixels:
            raise KeyError("start of range of pixels to read is out of range")

        if stop < 0 or stop > self._nr_pixels:
            raise KeyError("end of range of pixels to read is out of range")

        # Create output array
        pixels = np.ndarray(stop-start, dtype=self._dtype)

        # Check for zero size output
        if stop == start:
            if unyt is None:
                return pixels
            else:
                return unyt.unyt_array(pixels, self._units)

        # Determine which files to read
        nr_files = len(self._filenames)
        first_file = start // self._pix_per_file
        last_file = (stop-1) // self._pix_per_file
        if last_file >= nr_files:
            last_file = nr_files - 1

        # Loop over files
        output_offset = 0
        for file_nr in range(first_file, last_file+1):

            # Find offset to first pixel in this file
            file_offset = file_nr*self._pix_per_file

            # Open this file
            with self.open_file(self._filenames[file_nr]) as infile:

                # Find number of pixels in the file
                nr_pix_in_file = infile[self._map_name].shape[0]

                # Find range of pixels to read from this file
                i1 = start - file_offset
                if i1 < 0:
                    i1 = 0
                i2 = stop - file_offset
                if i2 > nr_pix_in_file:
                    i2 = nr_pix_in_file

                # Read pixels
                if i2 > i1:
                    nr_to_read = i2 - i1
                    pixels[output_offset:output_offset+nr_to_read] = infile[self._map_name][i1:i2]
                    output_offset += nr_to_read

        assert output_offset == (stop-start)

        if unyt is None:
            return pixels
        else:
            return unyt.unyt_array(pixels, self._units)

    def __getitem__(self, item):
        self._set_metadata()
        if isinstance(item, slice):
            start = item.start
            if start is None:
                start = 0
            if start < 0:
                start = self._nr_pixels + start
            stop = item.stop
            if stop is None:
                stop = self._nr_pixels
            if stop < 0:
                stop = self._nr_pixels + stop
            if stop < start:
                stop = start
            step = item.step
            if step is not None and step != 1:
                raise KeyError("Healpix map slices must be contiguous")
        elif item is Ellipsis:
            start = 0
            stop = self._nr_pixels
        else:
            raise KeyError("HealpixMap index must be range [x:y] or Ellipsis [...]")

        return self.read_pixels(start, stop)


class Shell(collections.abc.Mapping, LocalOrRemoteFile):
    """
    Dict-like container for all of the :class:`HealpixMap` instances
    associated with a lightcone shell. Subscripting a :class:`Shell`
    with a map name returns a :class:`HealpixMap`.

    :param basedir: directory containing the lightcone outputs
    :type  basedir: str
    :param basename: name of the subdirectory for this lightcone (e.g. ``lightcone0``)
    :type  basename: str
    :param shell_nr: index of the lightcone shell to read
    :type  shell_nr: int
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, basedir, basename, shell_nr, remote_dir=None):
        self.set_directory(remote_dir)

        # Find all files that make up this shell
        self.filenames = []
        file_nr = 0
        nr_files = 1
        while file_nr < nr_files:
            fname = ("%s/%s_shells/shell_%d/%s.shell_%d.%d.hdf5" %
                     (basedir, basename, shell_nr, basename, shell_nr, file_nr))
            self.filenames.append(fname)
            if file_nr == 0:
                with self.open_file(fname) as infile:
                    nr_files = int(infile["Shell"].attrs["nr_files_per_shell"][0])
                    length_unit_cgs = float(infile["Units"].attrs["Unit length in cgs (U_L)"][0])
                    self.comoving_inner_radius = infile["Shell"].attrs["comoving_inner_radius"][0]
                    self.comoving_outer_radius = infile["Shell"].attrs["comoving_outer_radius"][0]
                    if unyt is not None:
                        length_unit = unyt.cm * length_unit_cgs
                        self.comoving_inner_radius = unyt.unyt_quantity(self.comoving_inner_radius, units=length_unit)
                        self.comoving_outer_radius = unyt.unyt_quantity(self.comoving_outer_radius, units=length_unit)
                    self.map_names = []
                    for name in infile:
                        if "nside" in infile[name].attrs:
                            self.map_names.append(name)
            file_nr += 1

        # Create the map objects
        self._maps = {}
        for map_name in self.map_names:
            self._maps[map_name] = HealpixMap(self.filenames, map_name, self._remote_dir)

    def __getitem__(self, key):
        return self._maps[key]

    def __iter__(self):
        for name in self._maps:
            yield name

    def __len__(self):
        return len(self._maps)


class ShellArray(collections.abc.Sequence, LocalOrRemoteFile):
    """
    Sequence-like container for a set of lightcone shells. This class is
    the recommended way to read lightcone HEALPix maps.

    A :class:`ShellArray` can be indexed with an integer to return a :class:`Shell`.

    :param basedir: directory containing the lightcone outputs
    :type  basedir: str
    :param basename: name of the subdirectory for this lightcone (e.g. ``lightcone0``)
    :type  basename: str
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, basedir, basename, remote_dir=None):
        self.set_directory(remote_dir)

        # Get number of shells from the index file
        fname = basedir+"/"+basename+"_index.hdf5"
        with self.open_file(fname) as infile:
            self.nr_shells = infile["Lightcone"].attrs["nr_shells"][0]

        self.basedir = basedir
        self.basename = basename

        # Creating a Shell instance involves reading a file, so we don't
        # initialize shells until they're accessed.
        self._shell = []
        for shell_nr in range(self.nr_shells):
            self._shell.append(None)

    def __getitem__(self, index):
        # Initialize shell on access, if we didn't already
        if self._shell[index] is None:
            self._shell[index] = Shell(self.basedir, self.basename, index, self._remote_dir)
        return self._shell[index]

    def __len__(self):
        return len(self._shell)


def map_file_name(basedir, basename, shell_nr, file_nr):
    return ("%s/%s_shells/shell_%d/%s.shell_%d.%d.hdf5" %
            (basedir, basename, shell_nr, basename, shell_nr, file_nr))

def combine_healpix_maps(indir, basename, shell_nr, outdir):

    # Create the output file
    outname = map_file_name(outdir, basename, shell_nr, 0)
    os.makedirs(os.path.dirname(outname), exist_ok=True)
    outfile = h5py.File(outname, "w")
    print("Created output file: %s" % outname)

    # Open the first input file
    inname = map_file_name(indir, basename, shell_nr, 0)
    infile = h5py.File(inname, "r")
    print("Opened input file: %s" % inname)

    # Copy metadata
    infile.copy(source="Units", dest=outfile)
    infile.copy(source="InternalCodeUnits", dest=outfile)
    infile.copy(source="Shell", dest=outfile)
    outfile["Shell"].attrs["nr_files_per_shell"] = (1,)

    # Read the input unit information: this has the M, L, T etc units in cgs
    input_units_cgs = lightcone_io.units.read_cgs_units(infile)

    # Get list of datasets
    datasets = []
    for name in infile:
        if "nside" in infile[name].attrs:
            datasets.append(name)
            print("  found dataset %s" % name)

    # Determine total number of pixels
    number_of_pixels = infile[datasets[0]].attrs["number_of_pixels"]
    print("There are %d pixels per map" % number_of_pixels)

    # Create output datasets
    # Note: this will not preserve lossy compression filters
    for name in datasets:
        dset = infile[name]
        shape = (number_of_pixels,)
        outfile.create_dataset(name, shape=shape, dtype=dset.dtype,
                               chunks=dset.chunks, compression=dset.compression,
                               compression_opts=dset.compression_opts,
                               shuffle=dset.shuffle)
        for attr_name in infile[name].attrs:
            outfile[name].attrs[attr_name] = infile[name].attrs[attr_name]
        print("Created output dataset %s" % name)

        # Correct the unit metadata if necessary
        corrections = lightcone_io.units.correct_units(dset, name, input_units_cgs)
        if len(corrections) > 0:
            print("WARNING: modifying unit info for ", name)
        for attr_name, attr_value in corrections.items():
            print("  Setting attribute ", attr_name, " to ", attr_value)
            outfile[name].attrs[attr_name] = attr_value

    # Close the input file
    infile.close()

    # Copy data. Here we need to loop over the input files.
    offset = 0
    file_nr = 0
    while offset < number_of_pixels:

        # Open the next input file
        inname = map_file_name(indir, basename, shell_nr, file_nr)
        infile = h5py.File(inname, "r")
        print("Opened input file to copy pixel data: %s" % inname)

        # Copy pixel data from this file
        local_pixels = None
        for name in datasets:
            dset = infile[name]

            # Check number of pixels in this file
            if local_pixels is None:
                local_pixels = dset.shape[0]
            else:
                if local_pixels != dset.shape[0]:
                    raise Exception("All maps must be the same size!")

            # Copy the pixel data for this map
            outfile[name][offset:offset+local_pixels] = infile[name][:]

        # Next file
        offset += local_pixels
        file_nr += 1

    if offset != number_of_pixels:
        raise Exception("Number of pixels copied is wrong!")

    print("Finished merging files")


def fetch_pixels(indir, basename, shell_nr, dataset):

    max_per_fetch = 10*1024*1024

    # Loop over input files
    file_nr = 0
    total_read = 0
    while True:

        # Open the current file and dataset
        inname = map_file_name(indir, basename, shell_nr, file_nr)
        infile = h5py.File(inname, "r")
        dset = infile[dataset]
        if file_nr == 0:
            total_pixels = dset.attrs["number_of_pixels"]

        # Loop over and return pixels in this file
        for offset in range(0, dset.shape[0], max_per_fetch):
            i1 = offset
            i2 = offset + max_per_fetch
            if i2 > dset.shape[0]:
                i2 = dset.shape[0]
            assert (i2 > i1)
            yield dset[i1:i2]
            total_read += (i2 - i1)

        # Next file
        infile.close()
        file_nr += 1

        # Check if we're done
        assert total_read <= total_pixels
        if total_read == total_pixels:
            break


def compare_healpix_maps(indir1, indir2, basename, shell_nr, dataset):

    # Compute hash of first map
    map1 = fetch_pixels(indir1, basename, shell_nr, dataset)
    map1_hash = hashlib.sha256()
    for pixel_data in map1:
        map1_hash.update(pixel_data)
    map1_hash = map1_hash.digest()

    # Compute hash of second map
    map2 = fetch_pixels(indir2, basename, shell_nr, dataset)
    map2_hash = hashlib.sha256()
    for pixel_data in map2:
        map2_hash.update(pixel_data)
    map2_hash = map2_hash.digest()

    return (map1_hash == map2_hash)


def get_map_names(fname):
    map_names = []
    with h5py.File(fname, "r") as infile:
        for name in infile:
            if isinstance(infile[name], h5py.Dataset):
                map_names.append(name)
    return map_names



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'lightcone_io'
copyright = '2025, John Helly'
author = 'John Helly'

from importlib.metadata import version
release = version('lightcone_io')
version = '.'.join(release.split('.')[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx_autodoc_typehints',
    'sphinx.ext.intersphinx',
]

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# API documentation ordering
autodoc_member_order = 'bysource'

# This allows linking to sphinx docs of other projects
intersphinx_mapping = dict(
    unyt=("https://unyt.readthedocs.io/en/stable/", None),
    h5py=("https://docs.h5py.org/en/latest/", None),
    hdfstream=("https://hdfstream-python.readthedocs.io/en/latest/", None),
)



#!/bin/env python

import sys
import glob
import h5py
import numpy as np

def check_index_file(basedir, basename):

    print("Lightcone: ", basename)

    # Read the index
    filename="%s/%s_index.hdf5" % (basedir, basename)
    with h5py.File(filename, "r") as infile:
        nr_mpi_ranks = int(infile["Lightcone"].attrs["nr_mpi_ranks"])
        final_particle_file_on_rank = infile["Lightcone"].attrs["final_particle_file_on_rank"]
        nr_files_per_shell = int(infile["Lightcone"].attrs["nr_files_per_shell"])

    # Count particle files
    print("  Check number of particle files")
    fnames = glob.glob("%s/%s_particles/%s_0000.*.hdf5" % (basedir, basename, basename))
    if len(fnames) == 0:
        # No particle files: check that this is expected from the index file
        if np.all(final_particle_file_on_rank==-1):
            print("    OK: No particle files, as expected")
        else:
            raise Exception("ERROR: Missing particle files!")
    else:
        # Have some particle files. Check expected number
        if(np.all(final_particle_file_on_rank == 0)):
            print("    OK: one file per rank as expected")
        else:
            raise Exception("ERROR: should have one file per rank!")
        # Check number of files
        if len(fnames) == nr_mpi_ranks:
            print("    OK: have expected number of particle files")
        else:
            raise Exception("ERROR: unexpected number of particle files")

    print("  Check lightcone maps")
    if nr_files_per_shell != 1:
        raise Exception("ERROR: should have one file per shell!")
    else:
        print("    OK: one file per shell")


if __name__ == "__main__":

    basedir=sys.argv[1]
    nr_lightcones=int(sys.argv[2])

    for lightcone_nr in range(nr_lightcones):
        check_index_file(basedir, "lightcone%d" % lightcone_nr)



#!/bin/env python

import numpy as np
import healpy as hp
import unyt

from .units import units_from_attributes
from .utils import IndexedDatasetReader, SlicedDatasetReader, LocalOrRemoteFile


class HaloLightconeFile(LocalOrRemoteFile):
    """
    Class used to read a single file from the halo lightcone. Each
    file contains halos which were identified in a single snapshot of
    the simulation. Halos are sorted according to their pixel index in
    a low resolution HEALPix map so that regions on the sky can be
    extracted efficiently.

    If the name of the SOAP file which was used to generate the lightcone
    halo catalogue is supplied, then it will be possible to look up extra
    halo properties from SOAP.

    :param filename: Name of the halo lightcone file to open
    :type  filename: str
    :param soap_filename: Name of the SOAP output with extra halo properties
    :type  soap_filename: str
    :param remote_dir: remote directory containing the file, or None
    :type  remote_dir: hdfstream.RemoteDirectory, or None for local files
    """
    def __init__(self, filename, soap_filename=None, remote_dir=None):
        self.set_directory(remote_dir)
        self._file = self.open_direct(filename)
        if soap_filename is not None:
            self._soap_file = self.open_direct(soap_filename)
        else:
            self._soap_file = None
        self._num_halos_per_pixel = self._file["Index/NumHalosPerPixel"][...]
        self._first_halo_in_pixel = self._file["Index/FirstHaloInPixel"][...]
        self._nside = int(self._file["Index"].attrs["nside"])
        self._order = str(self._file["Index"].attrs["order"])
        self._props = None

    @property
    def properties(self):
        """
        Return a list of all halo property names in this file. These are the
        values which are valid to pass to the properties parameter of
        :py:meth:`read_halos_in_pixels` and :py:meth:`read_halos_in_radius`.

        Note that this does not include any additional properties which we
        may be able to read from SOAP.
        """
        if self._props is None:
            self._props = self._file["Lightcone"].attrs["property_names"]
        return self._props

    def get_pixels_in_radius(self, vector, radius):
        """
        Return indexes of all healpix pixels within radius of vector

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float

        :return: a numpy array of ints with the pixel indexes
        :rtype: numpy.ndarray
        """
        if self._order == "nest":
            nest = True
        elif self._order == "ring":
            nest = False
        else:
            raise RuntimeError("Invalid order parameter")
        pixels = hp.query_disc(self._nside, vector, radius, inclusive=True, nest=nest)
        pixels.sort()
        return pixels

    def read_halos_in_pixels(self, pixels, properties):
        """
        Read halos in the specified HEALPix pixels and return the
        requested halo properties in a dict of unyt arrays.

        If a SOAP filename was specifed then we'll check the SOAP file for any
        halo properties which are not present in the halo lightcone file.

        :param pixels: array of HEALPix pixel indexes to read, or None to read all
        :type  pixels: numpy.ndarray, or None
        :param properties: list of halo properties (i.e. HDF5 dataset names) to read
        :type  properties: list of str

        :return: dict of arrays with the halo properties
        :rtype:  dict of unyt.unyt_array
        """

        # Name of the dataset in the halo lightcone which specifies the
        # corresponding index in SOAP.
        soap_index_name = "InputHalos/SOAPIndex"

        # Determine which, if any, properties are to be read from SOAP and
        # which are to be read from the lighcone.
        lightcone_properties = []
        soap_properties = []
        for name in properties:
            if name in self._file:
                lightcone_properties.append(name)
            elif self._soap_file is not None and name in self._soap_file:
                soap_properties.append(name)
            else:
                raise KeyError(f"Unable to locate halo property: {name}")

        # If we're reading anything from SOAP we'll need the SOAP index of each
        # halo, so make sure it's in the list of things to read
        if len(soap_properties) > 0 and soap_index_name not in lightcone_properties:
            lightcone_properties.append(soap_index_name)

        if pixels is not None:
            # Determine ranges of halos to read
            if np.any(pixels[1:] <= pixels[:-1]):
                raise RuntimeError("Pixel indexes must be unique and ascending!")

            # Discard any selected pixels with no halos
            pixels = pixels[self._num_halos_per_pixel[pixels] > 0]

            # Compute offset to first halo and number of halos per pixel
            offsets = self._first_halo_in_pixel[pixels]
            counts  = self._num_halos_per_pixel[pixels]

        else:
            # We're reading all of the halos, so we just have one range to read
            offsets = np.asarray((0,), dtype=int)
            counts = np.asarray((np.sum(self._num_halos_per_pixel),), dtype=int)

        # Compute expected number of halos
        nr_halos = sum(counts)

        # Read halo properties from the lightcone
        result = {}
        reader = SlicedDatasetReader(offsets, counts)
        for name in lightcone_properties:
            dataset = self._file[name]
            units = units_from_attributes(dataset)
            result[name] = unyt.unyt_array(reader.read(dataset), units)

        # Now read extra properties from SOAP, if necessary
        if len(soap_properties) > 0:

            # Get the index of each selected lightcone halo
            soap_index = result[soap_index_name].value

            # Read these indexes from the SOAP datasets
            reader = IndexedDatasetReader(soap_index)
            for name in soap_properties:
                dataset = self._soap_file[name]
                units = units_from_attributes(dataset)
                result[name] = unyt.unyt_array(reader.read(dataset), units)

        # If the SOAP index was not requested, don't return it
        if soap_index_name in result and soap_index_name not in properties:
            del result[soap_index_name]

        return result

    def read_halos_in_radius(self, vector, radius, properties):
        """
        Read halos in an angular radius around the specified line of sight
        vector and return the requested halo properties in a dict of unyt
        arrays. May also return some halos slightly outside the radius
        because if we read a pixel we read all of the halos in it.

        :param vector: direction vector as an array of 3 floats
        :type  vector: numpy.ndarray
        :param radius: angular radius in radians
        :type  radius: float
        :param properties: list of halo properties (i.e. HDF5 dataset names) to read
        :type  properties: list of str

        :return: dict of arrays with the halo properties
        :rtype:  dict of unyt.unyt_array
        """
        pixels = self.get_pixels_in_radius(vector, radius)
        return self.read_halos_in_pixels(pixels, properties)

    def read_halos(self, properties):
        """
        Read all halos in this file and return the requested halo
        properties in a dict of unyt arrays.

        :param properties: list of halo properties (i.e. HDF5 dataset names) to read
        :type  properties: list of str

        :return: dict of arrays with the halo properties
        :rtype:  dict of unyt.unyt_array
        """
        return self.read_halos_in_pixels(None, properties)

    def __del__(self):
        """
        Close any open files
        """
        if self._file:
            self._file.close()
        if self._soap_file:
            self._soap_file.close()



#!/bin/env python

import h5py
import healpy as hp
import numpy as np
import unyt
import pytest

from lightcone_io import ShellArray, Shell, HealpixMap

basedir  = "tests/data/healpix_maps"
basename = "lightcone1"
shell_filename = "tests/data/healpix_maps/lightcone1_shells/shell_{shell_nr}/lightcone1.shell_{shell_nr}.0.hdf5"
maps = ("DarkMatterMass", "NeutrinoMass", "TotalMass")
nside = 16
nr_pixels = hp.nside2npix(nside)
nr_shells = 10


def test_open_shellarray(remote_dir):
    """
    Check we can open an array of healpix map shells
    """
    shells = ShellArray(basedir, basename, remote_dir=remote_dir)
    assert len(shells) == nr_shells
    for shell in shells:
        assert isinstance(shell, Shell)


def test_shell_info(remote_dir):
    """
    Check that shell metadata is correct
    """
    shells = ShellArray(basedir, basename, remote_dir=remote_dir)
    for shell_nr, shell in enumerate(shells):
        assert set(maps) == set(shell.keys())
        with h5py.File(shell_filename.format(shell_nr=shell_nr), "r") as infile:
            r_inner = infile["Shell"].attrs["comoving_inner_radius"][0]
            r_outer = infile["Shell"].attrs["comoving_outer_radius"][0]
        assert isinstance(shell.comoving_inner_radius, unyt.unyt_quantity)
        assert isinstance(shell.comoving_outer_radius, unyt.unyt_quantity)
        assert shell.comoving_inner_radius.value == r_inner
        assert shell.comoving_outer_radius.value == r_outer


def test_map_info(remote_dir):
    """
    Check that healpix map metadata is correct
    """
    shells = ShellArray(basedir, basename, remote_dir=remote_dir)
    shell_nr = 0
    shell = shells[shell_nr]
    with h5py.File(shell_filename.format(shell_nr=shell_nr), "r") as infile:
        for name in shell:
            map_data = shell[name]
            assert map_data.nside == nside
            cgs_conversion = infile[name].attrs["Conversion factor to CGS (not including cosmological corrections)"][0]
            assert np.isclose(cgs_conversion, shell[name].units.in_cgs().value)
            assert map_data.dtype == infile[name].dtype
            assert len(map_data) == infile[name].attrs["number_of_pixels"][0]


def test_full_map(remote_dir):
    """
    Read full maps and check pixel values against h5py
    """
    shells = ShellArray(basedir, basename, remote_dir=remote_dir)
    for shell_nr, shell in enumerate(shells):
        with h5py.File(shell_filename.format(shell_nr=shell_nr), "r") as infile:
            for name in shell:
                assert np.all(shell[name][...].value == infile[name][...])


# Generate some random test cases for reading pixel subsets
N = 50
rng = np.random.default_rng(seed=0)
map_index = rng.integers(0, len(maps), N)
shell_nr = rng.integers(0, nr_shells, N)
offset = rng.integers(0, nr_pixels, N)
length = rng.integers(0, nr_pixels, N)
test_cases = []
for i in range(N):
    test_cases.append((shell_nr[i], maps[map_index[i]], offset[i], min(length[i], nr_pixels-offset[i])))


@pytest.mark.parametrize("shell_nr,name,offset,length", test_cases)
def test_read_pixels(remote_dir, shell_nr, name, offset, length):
    """
    Try reading some pixels from a map and check against h5py
    """
    map_data = ShellArray(basedir, basename, remote_dir=remote_dir)[shell_nr][name]
    pixels = map_data[offset:offset+length].value
    with h5py.File(shell_filename.format(shell_nr=shell_nr), "r") as infile:
        assert np.all(pixels == infile[name][offset:offset+length])



#!/bin/env python

import re
import os
import os.path
import shutil
import sys
import gc

import h5py

import numpy as np
import healpy as hp
from mpi4py import MPI

import virgo.mpi.parallel_sort as ps

import lightcone_io.particle_metadata as lm
import lightcone_io.memory_use as memory_use

import h5py.h5t
import h5py.h5s


class LightconeSorter:

    def message(self, message):
        if self.comm.Get_rank()==0:
            print(message)
        memory_use.report_usage(self.comm)
            
    def __init__(self, basedir, basename, comm, types=None):
        
        self.comm = comm
        comm_size = comm.Get_size()
        comm_rank = comm.Get_rank()
        
        self.basedir = basedir
        self.basename = basename

        # Determine number of particles per file etc
        self.metadata = lm.LightconeMetadata(basedir, basename, comm)
        self.metadata.open_image()

        # Assign equally sized, contiguous ranges of particles to MPI ranks
        self.particle_offset = {}
        self.particle_count = {}
        for ptype in self.metadata.part_types:
            if self.metadata.nr_particles_total[ptype] > 0 and (types is None or ptype in types):
                nr_per_rank = self.metadata.nr_particles_total[ptype] // comm_size
                if nr_per_rank < 1:
                    raise Exception("Must have at least one particle per MPI rank!")
                self.particle_offset[ptype] = nr_per_rank * comm_rank
                self.particle_count[ptype] = nr_per_rank
                if comm_rank == comm_size-1:
                    self.particle_count[ptype] += (self.metadata.nr_particles_total[ptype] - nr_per_rank*comm_size)
                # Check total number of particles assigned is correct
                assert self.metadata.nr_particles_total[ptype] == self.comm.allreduce(self.particle_count[ptype])
            else:
                self.particle_count[ptype] = 0
                self.particle_offset[ptype] = 0

    def read_array(self, particle_type, property_name):
        """
        Read one particle property, distributed over all MPI ranks
        """

        # Allocate an array with the right data type and dimensions
        prop  = self.metadata.file[particle_type][property_name]
        dtype = prop.dtype.newbyteorder("=")
        shape = (self.particle_count[particle_type],)+prop.shape[1:]
        data  = np.ndarray(shape, dtype=dtype)

        # Find range of particles in each file
        offset_file = self.metadata.offset_file[particle_type]
        nr_particles_file = self.metadata.nr_particles_file[particle_type]

        # Loop over files
        output_offset = 0
        for file_nr in range(self.metadata.nr_files):
            
            # Find indexes i1:i2 of particles to read in this file
            i1 = self.particle_offset[particle_type] - offset_file[file_nr]
            i2 = i1 + self.particle_count[particle_type]
            if i1 < 0:
                i1 = 0
            if i2 > self.metadata.nr_particles_file[particle_type][file_nr]:
                i2 = self.metadata.nr_particles_file[particle_type][file_nr]
            
            # Read the data, if we have any particles to read from this file
            nread = i2-i1
            if nread > 0:
                with h5py.File(self.metadata.filenames[file_nr], "r") as infile:
                    data[output_offset:output_offset+nread,...] = infile[particle_type][property_name][i1:i2,...]
                output_offset += nread

        assert output_offset == self.particle_count[particle_type]
        return data

    def compute_bins(self, particle_type, nr_redshift_bins, nside, order,
                     redshift_first):
        """
        Decide on the redshift and healpix bins for this lightcone.

        Redshift bin edges are computed such that we have an approximately
        equal number of particles per bin.

        Returns an array of bins and a sorting index which puts the
        particles in order of which bin they belong to.
        """
        
        # HEALPix pixel ordering convention
        if order == "nest":
            nest=True
        elif order == "ring":
            nest=False
        else:
            raise Exception("Invalid order parameter")

        # Find which healpix pixel each particle is in
        self.message("  read coordinates")
        pos = self.read_array(particle_type, "Coordinates")
        self.message("  compute healpix pixel index")
        hp_bin_index = hp.pixelfunc.vec2pix(nside, pos[:,0], pos[:,1], pos[:,2], nest=nest)
        del pos
        
        # Get the redshift of each particle
        self.message("  compute redshift bin index")
        a = self.read_array(particle_type, "ExpansionFactors")
        assert len(a) == self.particle_count[particle_type]
        z = 1.0/a-1.0
        nr_particles_total = self.metadata.nr_particles_total[particle_type]

        # Get the full redshift range
        z_min = self.comm.allreduce(np.amin(z), op=MPI.MIN)
        z_max = self.comm.allreduce(np.amax(z), op=MPI.MAX)

        # Sort the redshifts
        z.sort()

        if nr_redshift_bins > 1:

            # Find global ranks of the particles at the bin edges
            # if we sort by redshift
            nr_per_bin = nr_particles_total // nr_redshift_bins
            rank = np.zeros(nr_redshift_bins-1, dtype=int)
            for i in range(nr_redshift_bins-1):
                rank[i] = (i+1)*nr_per_bin

            # Find the redshifts of these particles
            z_split, min_rank, max_rank = ps.find_splitting_points(z, rank, self.comm)

        elif nr_redshift_bins == 1:

            # Only one redshift bin, so no splitting points
            z_split = np.ndarray(0, dtype=float)

        else:
            raise Exception("Number of redshift bins must be positive!")
        del z

        # Make the full array of bin edges
        z_bins = np.ndarray(nr_redshift_bins+1)
        z_bins[0]    = z_min
        z_bins[1:-1] = z_split
        z_bins[-1]   = z_max

        # Find which redshift bin each particle is in
        z = 1.0/a-1.0 # Need un-sorted redshift here
        del a
        z_bin_index = np.searchsorted(z_bins, z, side="right")-1
        z_bin_index[z_bin_index==nr_redshift_bins] = nr_redshift_bins-1 # handle particle(s) with exactly z=z_max
        assert np.all((z >= z_bins[z_bin_index]) & (z <= z_bins[z_bin_index+1]))
        del z

        # Find number of pixels for this nside
        nr_pixels = hp.pixelfunc.nside2npix(nside)

        # Make an array of bins
        bin_dtype = np.dtype([("z_min", float),
                              ("z_max", float),
                              ("pixel", np.int64),
                              ("offset", np.int64),
                              ("length", np.int64)])
        bins = np.ndarray(nr_pixels*nr_redshift_bins, dtype=bin_dtype)
        offset = 0

        if redshift_first:
            # Sort bins by redshift, then by pixel within a redshift bin
            for bin_nr in range(nr_redshift_bins):
                for pixel_nr in range(nr_pixels):
                    bins[offset]["pixel"] = pixel_nr
                    bins[offset]["z_min"] = z_bins[bin_nr]
                    bins[offset]["z_max"] = z_bins[bin_nr+1]
                    bins[offset]["offset"] = 0
                    bins[offset]["length"] = 0
                    offset += 1
            bin_index = hp_bin_index + (z_bin_index * nr_pixels)
        else:
            # Sort bins by pixel, then by redshift within a pixel
            for pixel_nr in range(nr_pixels):
                for bin_nr in range(nr_redshift_bins):
                    bins[offset]["pixel"] = pixel_nr
                    bins[offset]["z_min"] = z_bins[bin_nr]
                    bins[offset]["z_max"] = z_bins[bin_nr+1]
                    bins[offset]["offset"] = 0
                    bins[offset]["length"] = 0
                    offset += 1
            bin_index = z_bin_index + (hp_bin_index * nr_redshift_bins)

        del z_bin_index
        del hp_bin_index

        # Count number of particles in each bin, summed over all MPI ranks
        bins["length"] = np.bincount(bin_index, minlength=len(bins))
        bins["length"] = self.comm.allreduce(bins["length"], op=MPI.SUM)

        # Compute offset to each bin if particles are sorted
        bins["offset"] = np.cumsum(bins["length"]) - bins["length"]

        # Generate a sorting index to put particles in order of bin
        self.message("  sort particles by bin index")
        sort_index = ps.parallel_sort(bin_index, comm=self.comm, return_index=True)

        # Return the array of bins and the sorting index
        return z_bins, bins, sort_index

    def write_sorted_lightcone(self, new_basedir, nr_redshift_bins, nside,
                               order="ring", redshift_first=True, lossy=True,
                               chunksize=0):
        """
        Write out a new lightcone with particles sorted by bin index
        """

        gc.collect()

        comm_rank = self.comm.Get_rank()
        comm_size = self.comm.Get_size()

        self.message("Copying index file")
        if comm_rank == 0:
            # Copy the file, if it doesn't already exist
            new_index_file_name = self.metadata.index_file_name(basedir=new_basedir)
            if not(os.path.exists(new_index_file_name)):
                os.makedirs(os.path.dirname(new_index_file_name), exist_ok=True)
                shutil.copyfile(self.metadata.index_file_name(), new_index_file_name)
            # Update number of files: will write one file per MPI rank
            with h5py.File(new_index_file_name, "r+") as new_index_file:
                new_index_file["Lightcone"].attrs["nr_mpi_ranks"] = (comm_size,)
                new_index_file["Lightcone"].attrs["final_particle_file_on_rank"] = np.zeros(comm_size, dtype=int)

        self.message("Creating output file")
        new_particle_file_name = self.metadata.particle_file_name(comm_rank, 0, new_basedir)
        os.makedirs(os.path.dirname(new_particle_file_name), exist_ok=True)
        outfile = h5py.File(new_particle_file_name, "w", libver=("v108", "latest"))
        outfile.create_group("Cells")

        with h5py.File(self.metadata.filenames[0], "r") as infile:
            # Copy metadata groups to the new output file
            infile.copy("Units", outfile)
            infile.copy("InternalCodeUnits", outfile)
            infile.copy("Lightcone", outfile)
            # Update some quantities
            outfile["Lightcone"].attrs["file_index"] = (0,)
            outfile["Lightcone"].attrs["mpi_rank"] = (comm_rank,)
            outfile["Lightcone"].attrs["nr_mpi_ranks"] = (comm_size,)
            del outfile["Lightcone"].attrs["expansion_factor"]

        # Loop over particle types
        for ptype in self.metadata.part_types:

            total_count = self.comm.allreduce(self.particle_count[ptype])
            if total_count > 0:

                self.message("Particle type %s (%d particles)" % (ptype, total_count))
                outfile.create_group(ptype)
                
                # Get sorting order and bins for this type
                z_bins, bins, sort_index = self.compute_bins(ptype, nr_redshift_bins, nside, order,
                                                             redshift_first)

                # Find offset to first particle in each file
                local_nr_particles = len(sort_index)
                local_particle_offset = self.comm.scan(local_nr_particles) - local_nr_particles 
                local_particle_offset = np.asarray(self.comm.allgather(local_particle_offset))
                local_nr_particles = np.asarray(self.comm.allgather(local_nr_particles))

                # Loop over quantities
                for name in self.metadata.properties[ptype]:

                    gc.collect()

                    # Find dataset template
                    dset_in = self.metadata.file[ptype][name]
                    
                    # Read the data for this particle property and sort into order of bin
                    self.message("  reading %s" % name)
                    data = self.read_array(ptype, name)
                    self.message("  sorting %s" % name)

                    if len(data.shape) == 1:
                        # Scalar quantity: just rearrange
                        ps.fetch_elements(data, sort_index, comm=self.comm, result=data)
                    elif len(data.shape) == 2:
                        # Vector quantity: save memory by rearranging one dimension at a time
                        for i in range(data.shape[1]):
                            col = data[:,i].copy()
                            ps.fetch_elements(col, sort_index, comm=self.comm, result=col)
                            data[:,i] = col
                            del col
                            gc.collect()
                    else:
                        raise Exception("Arrays with >2 dimensions not supported!")
                    self.message("  writing %s" % name)
                    dtype_id = dset_in.id.get_type()
                    dcpl_id = dset_in.id.get_create_plist()

                    # Disable lossy compression, if requested
                    if not(lossy):
                        try:
                            dcpl_id.remove_filter(h5py.h5z.FILTER_SCALEOFFSET)
                        except RuntimeError:
                            pass

                    # Override chunk size, if requested
                    if chunksize != 0:
                        chunks = (chunksize,)+(dset_in.shape[1:])
                        dcpl_id.set_chunk(chunks)

                    # Create the dataset
                    shape = (local_nr_particles[comm_rank],)+dset_in.shape[1:]
                    lm.create_dataset(outfile[ptype], name, dtype_id, shape, dcpl_id)
                    outfile[ptype][name][...] = data
                    del data
                    gc.collect()

                    # Add metadata to the array
                    dset_out = outfile[ptype][name]
                    for attr_name in dset_in.attrs:
                        dset_out.attrs[attr_name] = dset_in.attrs[attr_name]

                del sort_index
                gc.collect()

                # Update particle number for this type
                outfile["Lightcone"].attrs["cumulative_count_"+ptype] = (local_nr_particles[comm_rank],)

                # Write out binning information for this output
                cells = outfile["Cells"].create_group(ptype)
                cells["num_cells"] = len(bins)
                for name in bins.dtype.fields:
                    cells["cell_"+name] = bins[name]
                cells["first_particle_in_file"] = local_particle_offset
                cells["num_particles_in_file"]  = local_nr_particles
                cells["nside"] = nside
                cells["redshift_bins"] = z_bins
                cells["order"] = order
                cells["redshift_first"] = 1 if redshift_first else 0

        self.message("Done.")
        outfile.close()



#!/bin/env python

import os
import sys
import numpy as np
import h5py
import argparse
from mpi4py import MPI
import unyt

import virgo.mpi.parallel_hdf5 as phdf5
import virgo.mpi.parallel_sort as psort
from virgo.mpi.util import MPIArgumentParser
from virgo.util.partial_formatter import PartialFormatter
import virgo.formats.swift as swift

import lightcone_io.halo_catalogue as hc

comm = MPI.COMM_WORLD
comm_rank = comm.Get_rank()
comm_size = comm.Get_size()


def message(m):
    if comm_rank == 0:
        print(m)


def distributed_amax(arr, comm):
    """
    Find maximum over distributed array arr, bearing in mind that some
    MPI ranks might have zero elements.

    Returns None if all ranks have no elements.
    """
    if len(arr) == 0:
        local_max = None
    else:
        local_max = np.amax(arr)
    all_max = comm.allgather(local_max)
    global_max = None
    for m in all_max:
        if m is not None:
            if global_max is None:
                global_max = m
            elif m > global_max:
                global_max = m
    return global_max


def choose_bh_tracer(halo_index, snap_nr, final_snap_nr, snapshot_format,
                     membership_format, membership_cache, part_type,
                     NULL_BH_ID):
    """
    Find the ID of a suitable tracer particle for each subhalo
    Ideally we want to pick a black hole that exists at the next
    and previous snapshots.

    halo_index: halo index used in the membership files
    snap_nr: snapshot at which the subhalos exist
    final_snap_nr: final snapshot number in the simulation
    snapshot_format: format string for snapshot filenames
    membership_format: format string for group membership filenames
    membership_cache: stores previously read BH halo membership
    part_type: particle type to use to trace halos
    """

    # Ensure we have the BH IDs and halo membership for snapshots snap_nr-1,
    # snap_nr and snap_nr+1 (may be in membership_cache already)
    for sn in (snap_nr+1, snap_nr, snap_nr-1):
        if (sn >= 0) and (sn <= final_snap_nr) and (sn not in membership_cache):

            # Discard excess cache entries
            while len(membership_cache) > 2:
                max_snap_in_cache = max(membership_cache.keys())
                del membership_cache[max_snap_in_cache]

            # Create unit registry for this snapshot
            if comm_rank == 0:
                with h5py.File(snapshot_format.format(snap_nr=sn, file_nr=0), "r") as infile:
                    reg = swift.soap_unit_registry_from_snapshot(infile)
            else:
                reg = None
            reg = comm.bcast(reg)

            # Read in the black hole particle IDs and positions for this snapshot
            filenames = PartialFormatter().format(snapshot_format, snap_nr=sn, file_nr=None)            
            mf1 = phdf5.MultiFile(filenames, file_nr_attr=("Header","NumFilesPerSnapshot"), comm=comm)
            snap_bh_ids, snap_bh_pos = mf1.read((f"{part_type}/ParticleIDs", f"{part_type}/Coordinates"), unpack=True, read_attributes=True)
            
            # Check for the case where there are no BHs at this snapshot: MultiFile.read()
            # returns None if no ranks read any elements.
            if snap_bh_ids is None:
                membership_cache[sn] = (None, None, None, None, reg)
                continue

            # Add unit info to the positions
            pos_units = swift.soap_units_from_attributes(snap_bh_pos.attrs, reg)
            snap_bh_pos = unyt.unyt_array(snap_bh_pos, units=pos_units)
            
            # Read in the black hole particle halo membership
            filenames = PartialFormatter().format(membership_format, snap_nr=sn, file_nr=None)
            mf2 = phdf5.MultiFile(filenames, file_idx=mf1.all_file_indexes, comm=comm)
            (snap_bh_grnr, snap_bh_rank) = mf2.read((f"{part_type}/GroupNr_bound",
                                                     f"{part_type}/Rank_bound"), unpack=True)
            assert len(snap_bh_grnr) == len(snap_bh_ids)
            assert len(snap_bh_rank) == len(snap_bh_ids)

            # Add this snapshot to the cache
            membership_cache[sn] = (snap_bh_ids, snap_bh_grnr, snap_bh_rank, snap_bh_pos, reg)
            nr_bh_local = len(snap_bh_ids)
            nr_bh_tot = comm.allreduce(nr_bh_local)
            message(f"    Read {nr_bh_tot} BHs for snapshot {sn}")

    # Check if we have any black holes at this snapshot
    if membership_cache[snap_nr][0] is None:
        reg = membership_cache[snap_nr][4]
        tracer_bh_id = np.ndarray(len(halo_index), dtype=np.int64) * unyt.Unit("dimensionless", registry=reg)
        tracer_bh_id[:] = NULL_BH_ID
        tracer_bh_pos = -np.ones((len(halo_index),3), dtype=float) * unyt.Unit("a*snap_length", registry=reg)
        return tracer_bh_id, tracer_bh_pos

    # Get number of black holes at snapshot snap_nr
    nr_bh_local = len(membership_cache[snap_nr][0])
    nr_bh_tot = comm.allreduce(nr_bh_local)
    
    # Determine which black hole particles exist at the next snapshot
    bh_id_this = membership_cache[snap_nr][0]
    if snap_nr+1 <= final_snap_nr:
        bh_id_next = membership_cache[snap_nr+1][0]
    else:
        bh_id_next = None
    if bh_id_this is not None and bh_id_next is not None and part_type!="PartType1":    
        idx_at_next_snap = psort.parallel_match(bh_id_this, bh_id_next, comm=comm)
        exists_at_next_snap = idx_at_next_snap >= 0
    else:
        exists_at_next_snap = np.ones(nr_bh_local, dtype=bool)
    nr_existing_next = comm.allreduce(np.sum(exists_at_next_snap))
    message(f"    Number of BHs which exist at the next snapshot = {nr_existing_next}")

    # Determine which black hole particles exist at the previous snapshot
    bh_id_this = membership_cache[snap_nr][0]
    if snap_nr-1 >= 0:
        bh_id_prev = membership_cache[snap_nr-1][0]
    else:
        bh_id_prev = None
    if bh_id_this is not None and bh_id_prev is not None and part_type!="PartType1":
        idx_at_prev_snap = psort.parallel_match(bh_id_this, bh_id_prev, comm=comm)
        exists_at_prev_snap = idx_at_prev_snap >= 0
    else:
        exists_at_prev_snap = np.ones(nr_bh_local, dtype=bool)
    nr_existing_prev = comm.allreduce(np.sum(exists_at_prev_snap))
    message(f"    Number of BHs which exist at the previous snapshot = {nr_existing_prev}")

    # Assign priorities to the BH particles. In descending order of importance:
    # 1. Should exist at the next timestep
    # 2. Should exist at the previous timestep
    # 3. Should be tightly bound
    rank_at_this_snap = membership_cache[snap_nr][2]
    max_rank = distributed_amax(rank_at_this_snap, comm)
    if max_rank is None:
        max_rank = 1
    bh_priority = (max_rank - rank_at_this_snap).astype(np.int64) # Low rank = high priority
    assert np.all(bh_priority>=0)
    bh_priority += (max_rank+1)*exists_at_prev_snap               # Boost priority if exists at snap_nr-1
    bh_priority += 2*(max_rank+1)*exists_at_next_snap             # Boost priority more if exists at snap_nr+1
    bh_id   = membership_cache[snap_nr][0]
    bh_grnr = membership_cache[snap_nr][1]
    bh_pos = membership_cache[snap_nr][3]
    message("    BH priorities assigned")

    # Discard BHs which are not in halos
    keep = bh_grnr >= 0
    bh_priority = bh_priority[keep]
    bh_id = bh_id[keep]
    bh_grnr = bh_grnr[keep]
    bh_pos = bh_pos[keep,:]

    # Sort BH particles by halo and then by priority within a halo
    max_priority = distributed_amax(bh_priority, comm)
    if max_priority is None:
        max_priority = 1
    sort_key = bh_grnr * (max_priority+1) + bh_priority
    order = psort.parallel_sort(sort_key, return_index=True, comm=comm)
    del sort_key
    bh_priority = psort.fetch_elements(bh_priority, order, comm=comm)
    bh_id = psort.fetch_elements(bh_id, order, comm=comm)
    bh_grnr =  psort.fetch_elements(bh_grnr, order, comm=comm)
    bh_pos = psort.fetch_elements(bh_pos, order, comm=comm)
    del order
    message("    Sorted BH particles by priority")

    # Now we need to discard all but the last (i.e. highest priority) particle in each halo.
    # Discard any particle which is in the same halo as the next particle on the same rank.
    keep = np.ones(len(bh_id), dtype=bool)
    keep[:-1] = bh_grnr[:-1] != bh_grnr[1:]
    
    # The last particle on each rank needs special treatment:
    # We need to know the group index of the first particle on the next rank
    # which has particles. Find the first group number on every rank.
    if len(bh_grnr) > 0:
        first_grnr = bh_grnr[0]
    else:
        first_grnr = None
    first_grnr = comm.allgather(first_grnr)
    
    # Ranks other than the last need to determine if their last particle
    # is in the same group as the first particle on a later rank.
    if comm_rank < (comm_size-1) and len(bh_grnr) > 0:
        first_grnr_next_rank = None
        for i in range(comm_rank+1, comm_size):
            if first_grnr[i] is not None:
                first_grnr_next_rank = first_grnr[i]
                break
        if first_grnr_next_rank is not None:
            if bh_grnr[-1] == first_grnr_next_rank:
                # A later rank has a particle in the same halo as our last particle.
                # The later particle will have a higher priority so discard this one.
                keep[-1] = False

    # Discard particles which are not the highest priority in their halo
    bh_id = bh_id[keep]
    bh_grnr = bh_grnr[keep]
    bh_pos = bh_pos[keep,:]
    del bh_priority
    message("    Discarded low priority BH particles")

    # Now that we only have the highest priority particle in each halo,
    # for each halo find the 0-1 BH particles which belong to that halo
    # and look up their IDs.
    ptr = psort.parallel_match(halo_index, bh_grnr, comm=comm)
    nr_groups_matched = comm.allreduce(np.sum(ptr>=0))
    nr_groups_total = comm.allreduce(len(ptr))
    fraction_matched = nr_groups_matched / nr_groups_total
    message(f"    Matched fraction {fraction_matched:.2f} of halos to BHs")

    # Fetch the IDs and positions of the matched black holes. Return ID=NULL_BH_ID where there's no match.
    tracer_bh_id = np.ndarray(len(halo_index), dtype=bh_id.dtype)
    tracer_bh_id[:] = NULL_BH_ID
    tracer_bh_id[ptr>=0] = psort.fetch_elements(bh_id, ptr[ptr>=0], comm=comm)
    tracer_bh_pos = np.zeros((len(halo_index),3), dtype=bh_pos.dtype) * bh_pos.units
    tracer_bh_pos[ptr>=0,:] = psort.fetch_elements(bh_pos, ptr[ptr>=0], comm=comm)

    # As a consistency check, fetch the group number of the matched particles:
    # Each matched particle should belong to the group it was matched to.
    tracer_bh_grnr = np.ndarray(len(halo_index), dtype=bh_grnr.dtype)
    tracer_bh_grnr[:] = -1
    tracer_bh_grnr[ptr>=0] = psort.fetch_elements(bh_grnr, ptr[ptr>=0], comm=comm)
    assert np.all(tracer_bh_grnr[ptr>=0] == halo_index[ptr>=0])
    
    return tracer_bh_id, tracer_bh_pos


def test_choose_bh_tracers():
    """
    Identify BH particles to use to place halos on the lightcone
    """
    parser = MPIArgumentParser(comm, description='Determine tracer particle for each halo.')
    parser.add_argument('halo_format', help='Format string for halo catalogue files')
    parser.add_argument('snapshot_format', help='Format string for snapshot filenames')
    parser.add_argument('membership_format', help='Format string for group membership filenames')
    parser.add_argument('first_snap', type=int, help='Index of the first snapshot to use')
    parser.add_argument('last_snap', type=int, help='Index of the last snapshot to use')
    parser.add_argument('snap_nr', type=int, help='Snapshot number to process')
    parser.add_argument('output_file', help='Where to write the output')
    args = parser.parse_args()

    to_read = ("InputHalos/cofp", "InputHalos/index")
    
    message("Reading halo catalogue")
    halo_cat = hc.HaloCatalogue(args.halo_format, args.first_snap, args.last_snap)
    halo_data = halo_cat.read(args.snap_nr, to_read)

    message("Finding tracers")
    membership_cache = {}
    tracer_id, tracer_pos = choose_bh_tracer(halo_data["InputHalos/index"], args.snap_nr, args.last_snap,
                                             args.snapshot_format, args.membership_format, membership_cache)
    
    message("Writing results")
    with h5py.File(args.output_file, "w", driver="mpio", comm=comm) as f:
        phdf5.collective_write(f, "tracer_id", tracer_id, comm)

    comm.barrier()
    message("Done.")


if __name__ == "__main__":
    test_choose_bh_tracers()



#!/bin/env python
#
# This script is used to generate a tiny test SOAP input which can be used
# to make a similarly tiny lightcone halo catalogue.
#
# We random sample the SOAP halos and discard most of their properties. We need
# at least "InputHalos/HaloCentre" and "InputHalos/HaloCatalogueIndex". The
# HBT TrackId is also useful as a cross check.
#
# Need to sample all SOAP outputs for the halo lightcone code to run but we'll
# only keep one (and it's corresponding lightcone slice) in the repository.
#
# The slurm batch script in this directory was used to process the sampled
# SOAP catalogue into a halo lightcone file.
#

import h5py
import numpy as np


def sample_soap(snap_nr):

    print(f"Snapshot {snap_nr:04d}")

    # List of SOAP properties to keep
    to_keep = [
        "InputHalos/HaloCentre",
        "InputHalos/HaloCatalogueIndex",
        "InputHalos/HBTplus/TrackId",
        "SO/200_crit/TotalMass",
    ]

    # Location of the input catalogue
    input_filename = f"/cosma8/data/dp004/flamingo/Runs/L1000N1800/HYDRO_FIDUCIAL/SOAP-HBT/halo_properties_{snap_nr:04d}.hdf5"

    # Sampling rate, to keep the file small enough to commit to the git repo.
    # This should yield ~2,000 halos.
    sample_rate = 0.0005874 / 5

    # Location of the output
    output_filename = f"/cosma8/data/dp004/jch/halo_properties_{snap_nr:04d}.hdf5"

    # Create the output file
    outfile = h5py.File(output_filename, "w")

    # Open the input file
    infile = h5py.File(input_filename, "r")

    # Copy metadata groups
    for name in ("Cosmology", "Units", "PhysicalConstants"):
        infile.copy(infile[name], outfile, name)

    # Random sample the halos
    rng = np.random.default_rng(seed=0)
    nr_halos = int(infile["Header"].attrs["NumSubhalos_ThisFile"][0])
    keep = rng.random(nr_halos) < sample_rate
    print(f"Number of halos to keep = {sum(keep)}")

    # Copy over the halo data we need
    for name in to_keep:
        print(f"Copying: {name}")
        data = infile[name][...][keep,...]
        outfile.require_group(name.rsplit("/", 1)[0])
        outfile.create_dataset(name, data=data, compression="gzip", compression_opts=9)
        for attr_name, attr_val in infile[name].attrs.items():
            outfile[name].attrs[attr_name] = attr_val

if __name__ == "__main__":

    for snap_nr in range(78):
        sample_soap(snap_nr)



#!/bin/env python
#
# This script uses the lightcone particle data to create a new HEALPix map.
# Parallelised with MPI.

import healpy as hp
import numpy as np
import h5py
import unyt

import lightcone_io.particle_reader as pr
import lightcone_io.kernel as kernel

import virgo.mpi.parallel_sort as psort
import virgo.mpi.parallel_hdf5 as phdf5

from mpi4py import MPI
comm = MPI.COMM_WORLD
comm_size = comm.Get_size()
comm_rank = comm.Get_rank()

# May want to write one stderr file per MPI rank for debugging
#import sys
#sys.stderr = open(f"./stderr.{comm_rank}.log", "w")

projected_kernel = kernel.ProjectedKernel()

def native_endian(arr):
    return arr.astype(arr.dtype.newbyteorder("="))


def find_angular_smoothing_length(pos, hsml):
    """
    Return smoothing length in radians of the supplied particles,
    using small angle approximation for consistency with SWIFT
    """
    dist = np.sqrt(np.sum(pos**2, axis=1, dtype=float))
    small_angle = dist > 10.0*hsml
    large_angle = (small_angle==False)
    angular_smoothing_length = np.ndarray(pos.shape[0], dtype=float)
    ratio = (hsml/dist).to(unyt.dimensionless)
    angular_smoothing_length[small_angle] = ratio[small_angle]
    angular_smoothing_length[large_angle] = np.arctan(ratio[large_angle])
    return angular_smoothing_length


def explode_particle(nside, part_pos, part_val, angular_smoothing_length):
    """
    Given a particle's position vector and angular radius,
    return indexes of the pixels it will update and the values
    to add to the pixels.

    part_pos: particle's position vector
    part_val: particle's contribution to the map
    angular_smoothing_length: smoothing length in radians

    Returns

    pix_index: array of indexes of the pixels to update
    pix_val: array of values to add to the pixels
    """

    # Normalize position vector
    part_pos = part_pos / np.sqrt(np.sum(part_pos**2, dtype=float))

    # Find radius containing the pixels to update
    angular_search_radius = angular_smoothing_length*kernel.kernel_gamma
            
    # Get pixel indexes to update
    pix_index = hp.query_disc(nside, part_pos, angular_search_radius)
    assert len(pix_index) >= 1

    # For each pixel, find angle between pixel centre and the particle
    pix_vec_x, pix_vec_y, pix_vec_z = hp.pixelfunc.pix2vec(nside, pix_index)
    dp = part_pos[0]*pix_vec_x + part_pos[1]*pix_vec_y + part_pos[2]*pix_vec_z
    dp = np.clip(dp, a_min=None, a_max=1.0)
    pix_angle = np.arccos(dp)

    # Evaluate the projected kernel for each pixel
    pix_weight = projected_kernel(pix_angle/angular_smoothing_length)

    # Normalize weights so that sum is one
    pix_val = part_val * pix_weight / np.sum(pix_weight, dtype=float)

    return pix_index, pix_val


def exchange_particles(part_dest, arrays):

    # Sort arrays by destination
    send_arrays = []
    order = np.argsort(part_dest)
    for array in arrays:
        send_arrays.append(native_endian(array[order,...]))
    
    # Count particles to send to and receive from each rank
    send_count = np.bincount(part_dest, minlength=comm_size)
    recv_count = np.zeros_like(send_count)
    comm.Alltoall(send_count, recv_count)
    send_offset = np.cumsum(send_count) - send_count
    recv_offset = np.cumsum(recv_count) - recv_count
    nr_recv_tot = np.sum(recv_count)

    # Allocate receive buffers
    recv_arrays = []
    for array in send_arrays:
        shape = list(array.shape)
        shape[0] = nr_recv_tot
        recv_arrays.append(np.empty_like(array, shape=shape))

    # Exchange particles
    for sendbuf, recvbuf in zip(send_arrays, recv_arrays):
        if len(sendbuf.shape) == 1:
            # 1D array
            psort.my_alltoallv(sendbuf, send_count, send_offset,
                               recvbuf, recv_count, recv_offset,
                               comm=comm)
        elif len(sendbuf.shape) == 2:
            # 2D array: flatten before sending
            ndims = sendbuf.shape[1]
            psort.my_alltoallv(sendbuf.ravel(), ndims*send_count, ndims*send_offset,
                               recvbuf.ravel(), ndims*recv_count, ndims*recv_offset,
                               comm=comm)
        else:
            raise RuntimeError("Can only exchange arrays with 1 or 2 dimensions")

    return recv_arrays


def create_empty_array(arr, comm):
    """
    Ensure that if arr is None on some ranks it is replaced with a zero
    sized array of the same dtype and units as the other ranks.
    """
    has_units = comm.allreduce(hasattr(arr, "units"), op=MPI.MAX)
    local_dtype = None if arr is None else arr.dtype
    if has_units:
        local_units = None if arr is None else arr.units
    else:
        local_units = None
    local_shape = None if arr is None else arr.shape[1:]
    dtypes = comm.allgather(local_dtype)
    units  = comm.allgather(local_units)
    shapes = comm.allgather(local_shape)
    if arr is None:
        for dt, u, s in zip(dtypes, units, shapes):
            if dt is not None:
                arr = np.ndarray((0,)+s, dtype=dt)
                if has_units:
                    arr = unyt.unyt_array(arr, units=u, dtype=dt)
                return arr
        raise ValueError("All input arrays are None in create_empty_array()!")
    else:
        return arr


def message(m):
    comm.barrier()
    if comm_rank == 0:
        print(m)


def distribute_pixels(comm, nside):
    """
    Decide how to assign HEALPix pixels to MPI ranks.

    Here we assume the ring ordering scheme so that the minimum
    colatitude of a rank is the colatitude of its first pixel
    and the maximum colatitude of the rank is the colatitude of
    its last pixel, since all pixels in a ring have the same
    colatitude.

    We put equal numbers of pixels on each rank except that any
    leftovers are assigned to the last rank.

    Returns -
    
    nr_total_pixels: total number of pixels in the map
    nr_local_pixels: number of pixels on each rank
    local_offset:    offset between local and global pixel indexes
    min_theta:       minimum co-latitude theta on each rank
    max_theta:       minimum co-latitude theta on each rank
    """

    # Find number of pixels on each rank
    comm_rank = comm.Get_rank()
    comm_size = comm.Get_size()
    nr_total_pixels = hp.pixelfunc.nside2npix(nside)
    nr_local_pixels = nr_total_pixels // comm_size
    local_offset = nr_local_pixels * comm_rank

    # Find range of pixels on each rank
    first_pixel = np.arange(comm_size, dtype=int) * nr_local_pixels
    last_pixel = first_pixel + nr_local_pixels - 1
    # Any extra pixels go on the last rank
    last_pixel[-1] = nr_total_pixels - 1
    nr_local_pixels = last_pixel[comm_rank] - first_pixel[comm_rank] + 1

    # Find range in colatitude on each rank:
    # Here we make an array with the value of theta at the boundary between
    # ranks, with com_size+1 elements. The first and last elements are 0 and 2pi.
    theta_boundary = np.ndarray(comm_size+1, dtype=float)
    theta_boundary[0] = 0.0
    theta_boundary[-1] = np.pi
    for i in range(0,comm_size-1):
        theta1, phi = hp.pixelfunc.pix2ang(nside, last_pixel[i])
        theta2, phi = hp.pixelfunc.pix2ang(nside, first_pixel[i+1])
        theta_boundary[i+1] = 0.5*(theta1+theta2)

    return nr_total_pixels, nr_local_pixels, local_offset, theta_boundary


def make_sky_map(input_filename, ptype, property_names, particle_value_function,
                      zmin, zmax, nside, vector = None, radius = None, smooth=True, progress=False):
    """
    Make a new HEALPix map from lightcone particle data

    input_filename: name of one file from the particle lightcone output
    ptype         : which particle type to use
    property_names: list of particle properties to read in
    particle_value_function: function which returns the quantity to map,
                    given a dict containing the arrays specified by property_names
    zmin          : minimum redshift to use
    zmax          : maximum redshift to use
    nside         : HEALPix resolution parameter
    smooth        : whether to smooth the map
    """
    
    if progress and comm_rank == 0:
        from tqdm import tqdm
        progress_bar = tqdm
    else:
        progress_bar = lambda x: x

    # Ensure property_names list contains coordinates and smoothing lengths
    property_names = list(property_names)
    if "Coordinates" not in property_names:
        property_names.append("Coordinates")
    if smooth and ("SmoothingLengths" not in property_names):
        property_names.append("SmoothingLengths")

    # Open the lightcone
    lightcone = pr.IndexedLightcone(input_filename, comm=comm)

    # Create an empty HEALPix map, distributed over MPI ranks.
    # Here we assume we can put equal sized chunks of the map on each rank.
    nr_total_pixels, nr_local_pixels, local_offset, theta_boundary = distribute_pixels(comm, nside)
    max_pixrad = hp.pixelfunc.max_pixrad(nside)
    message(f"Total number of pixels = {nr_total_pixels}")
    
    # Will read the full sky within the redshift range
    redshift_range = (zmin, zmax)

    # Determine number of particles to read on this MPI rank
    nr_particles_local = lightcone[ptype].count_particles(redshift_range=redshift_range,
                                                          vector=vector, radius=radius,)
    nr_particles_total = comm.allreduce(nr_particles_local)
    message(f"Total number of particles in selected cells = {nr_particles_total}")

    # Read in the particle data
    particle_data = lightcone[ptype].read_exact(property_names, vector, radius, redshift_range)
    nr_parts_tot = comm.allreduce(particle_data["Coordinates"].shape[0])
    message(f"Read in {nr_parts_tot} particles")

    # Find the particle positions and smoothing lengths
    part_pos_send = particle_data["Coordinates"]
    if smooth:
        part_hsml_send = find_angular_smoothing_length(part_pos_send, particle_data["SmoothingLengths"])
    else:
        part_hsml_send = np.zeros(part_pos_send.shape[0], dtype=float)

    # Find the quantities to add to the map
    part_val_send = unyt.unyt_array(particle_value_function(particle_data), dtype=float)
    val_total_global = comm.allreduce(np.sum(part_val_send, dtype=float))

    # Find units of the quantity which we're mapping
    map_units = part_val_send.units
    all_map_units = comm.allgather(map_units)
    for unit in all_map_units:
        if unit != map_units:
            raise RuntimeError("Quantity to map needs to have the same units on all MPI ranks!")

    # No longer need the particle data
    del particle_data

    # Determine range of colatitudes each particle will update.
    # Smoothing kernel drops to zero at kernel_gamma * smoothing length.
    # Note that particles with radius < max_pixrad can still update
    # pixels up to max_pixrad away because they update whatever pixel
    # they are in.
    radius = np.maximum(kernel.kernel_gamma*part_hsml_send, max_pixrad) # Might update pixels with centres within this radius
    theta, phi = hp.pixelfunc.vec2ang(part_pos_send)
    part_min_theta = np.clip(theta-radius, 0.0, np.pi) # Minimum central theta of pixels each particle might update
    part_max_theta = np.clip(theta+radius, 0.0, np.pi) # Maximum central theta of pixels each particle might update

    # Determine what range of MPI ranks each particle needs to be sent to
    part_first_rank = np.searchsorted(theta_boundary, part_min_theta, side="left") - 1
    part_first_rank = np.clip(part_first_rank, 0, comm_size-1)
    part_last_rank  = np.searchsorted(theta_boundary, part_max_theta, side="right") - 1
    part_last_rank  = np.clip(part_last_rank, 0, comm_size-1)
    assert np.all(theta_boundary[part_first_rank]  <= part_min_theta)
    assert np.all(theta_boundary[part_last_rank+1] >= part_max_theta)

    # Determine how many ranks each particle needs to be sent to
    nr_copies = part_last_rank - part_first_rank + 1
    assert np.all(nr_copies>=1) and np.all(nr_copies<=comm_size)

    # Duplicate the particles
    nr_parts = part_pos_send.shape[0]
    index = np.repeat(np.arange(nr_parts, dtype=int), nr_copies)
    part_pos_send  = part_pos_send[index,...]
    part_val_send  = part_val_send[index]
    part_hsml_send = part_hsml_send[index]
    del index

    # Determine destination rank for each particle copy
    nr_parts = np.sum(nr_copies)                # Total number of copied particles
    offset = np.cumsum(nr_copies) - nr_copies   # Offset to first copy of each particle in array of copies
    part_dest = -np.ones(nr_parts, dtype=int) # Destination rank for each copied particle
    for pfr, nrc, off in zip(part_first_rank, nr_copies, offset):
        part_dest[off:off+nrc] = np.arange(pfr, pfr+nrc, dtype=int)
    assert np.all(part_dest >=0) & np.all(part_dest<comm_size)
    message("Computed destination rank(s) for each particle")

    # Tidy up
    del offset
    del part_first_rank
    del part_last_rank
    del nr_copies

    # Copy particles to their destinations
    part_pos_recv, part_val_recv, part_hsml_recv = (
        exchange_particles(part_dest, (part_pos_send, part_val_send, part_hsml_send)))

    # Free send buffers
    del part_pos_send
    del part_val_send
    del part_hsml_send
    del part_dest

    # Allocate the output map
    map_data = unyt.unyt_array(np.zeros(nr_local_pixels, dtype=float), units=map_units)

    # Will use unit-less views to carry out the map update to minimize overhead
    map_view = map_data.ndarray_view()
    part_pos_recv_view = part_pos_recv.ndarray_view()
    part_val_recv_view = part_val_recv.ndarray_view()

    # Now each MPI rank has copies of all particles which affect its local
    # pixels. Process any particles which update single pixels.
    single_pixel = part_hsml_recv*kernel.kernel_gamma < max_pixrad
    nr_single_pixel = comm.allreduce(np.sum(single_pixel))
    local_pix_index = hp.pixelfunc.vec2pix(nside, 
                                           part_pos_recv_view[single_pixel, 0],
                                           part_pos_recv_view[single_pixel, 1],
                                           part_pos_recv_view[single_pixel, 2]) - local_offset
    local = (local_pix_index >=0) & (local_pix_index < nr_local_pixels)
    np.add.at(map_view, local_pix_index[local], part_val_recv_view[single_pixel][local])
    del part_pos_recv_view
    del part_val_recv_view
    message(f"Applied {nr_single_pixel} single pixel updates")
    
    # Discard single pixel particles
    part_hsml_recv = part_hsml_recv[single_pixel==False]
    part_val_recv  = part_val_recv[single_pixel==False]
    part_pos_recv  = part_pos_recv[single_pixel==False,:]

    # Apply updates from particles which cover multiple pixels.
    #
    # Here we use ndarray views of the unyt part_val array and the map to avoid
    # any unit handling overhead. No conversion is necessary because arrays are 
    # in the same units.
    #
    # We only use the direction of the position vectors, so their units can
    # be ignored too.
    #
    nr_parts = len(part_val_recv)
    nr_parts_tot = comm.allreduce(nr_parts)
    part_pos_view  = part_pos_recv.ndarray_view()
    part_val_view  = part_val_recv.ndarray_view()
    for part_nr in progress_bar(range(nr_parts)):
        pix_index, pix_val = explode_particle(nside, part_pos_view[part_nr,:], part_val_view[part_nr], part_hsml_recv[part_nr])
        local_pix_index = pix_index - local_offset
        local = (local_pix_index >=0) & (local_pix_index < nr_local_pixels)
        # Don't need to use np.add.at here because pixel indexes are unique
        map_view[local_pix_index[local]] += pix_val[local]
    message(f"Applied {nr_parts_tot} multi-pixel updates")

    # Sanity check:
    # Sum over the map should equal sum of values to be accumulated to the map.
    map_sum = comm.allreduce(np.sum(map_data))
    ratio = map_sum / val_total_global
    message(f"Ratio (map sum / total values to add to map) = {ratio} (should be 1.0)")

    return map_data



#!/bin/env python

import healpy as hp
import numpy as np
import unyt
import pytest

from lightcone_io.halo_reader import HaloLightconeFile
from lightcone_io.utils import match

halo_lightcone_filename = "./tests/data/halo_lightcone/lightcone_halos_0070.hdf5"
soap_filename = "./tests/data/halo_lightcone/halo_properties_0070.hdf5"

# Quantities which should exist in the halo lightcone file
halo_properties = (
    "InputHalos/HaloCatalogueIndex",
    "InputHalos/HaloCentre",
    "InputHalos/SOAPIndex",
    "InputHalos/HBTplus/TrackId",
    "Lightcone/HaloCentre",
    "Lightcone/Redshift",
    "Lightcone/SnapshotNumber",
)

# Extra quantities we can get from SOAP
soap_properties = (
    "SO/200_crit/TotalMass",
)


def test_properties(remote_dir):
    """
    Check that the file has the expected set of properties
    """
    halos = HaloLightconeFile(halo_lightcone_filename, remote_dir=remote_dir)
    assert set(halo_properties) == set(halos.properties)


def test_read_everything(remote_dir):
    """
    Try reading all properties for all halos
    """
    # Read the file
    halos = HaloLightconeFile(halo_lightcone_filename, remote_dir=remote_dir)
    data = halos.read_halos(halo_properties)

    # Check values against reading the file directly
    with halos.open_file(halo_lightcone_filename) as infile:
        for name in halo_properties:
            assert isinstance(data[name], unyt.unyt_array)
            assert np.all(data[name].value == infile[name][...])


def test_read_everything_with_soap(remote_dir):
    """
    Try reading all properties for all halos. This version reads m200c from
    the SOAP file too.
    """
    # Open the file
    halos = HaloLightconeFile(halo_lightcone_filename, soap_filename=soap_filename, remote_dir=remote_dir)

    # Read all properties
    data = halos.read_halos(halo_properties+soap_properties)

    # Check values against halo lightcone file
    with halos.open_file(halo_lightcone_filename) as infile:
        for name in halo_properties:
            assert isinstance(data[name], unyt.unyt_array)
            assert np.all(data[name].value == infile[name][...])

    # Check values against SOAP
    soap_index = data["InputHalos/SOAPIndex"].value
    with halos.open_file(soap_filename) as infile:
        for name in soap_properties:
            assert isinstance(data[name], unyt.unyt_array)
            assert np.all(data[name].value == infile[name][...][soap_index,...])


def halo_id(pos):
    """
    Generate unique IDs for the halos in the lightcone, based on positions
    """
    nr_halos = pos.shape[0]
    halo_id = np.ndarray(nr_halos, dtype=np.uint64)
    for i in range(nr_halos):
        halo_id[i] = pos[i,:].astype(np.float64).view(np.uint64).sum()
    # As long as these happen to be unique for the test dataset, we're ok!
    assert len(halo_id) == len(np.unique(halo_id))
    return halo_id


def try_read_radius(vector, radius, properties, remote_dir):
    """
    Read halos in the specified radius about a vector and check that we get
    all of the halos within the radius.
    """
    # Read the specified halos and assign unique IDs to them
    partial_halos = HaloLightconeFile(halo_lightcone_filename, remote_dir=remote_dir)
    partial_data = partial_halos.read_halos_in_radius(vector, radius, halo_properties)
    partial_halo_ids = halo_id(partial_data["Lightcone/HaloCentre"])

    # Read all halos and assign unique IDs to them
    with partial_halos.open_file(halo_lightcone_filename) as infile:
        full_data = {}
        for name in partial_data:
            full_data[name] = infile[name][...]
        full_halo_ids = halo_id(full_data["Lightcone/HaloCentre"])

    # Identify which halos in the full set were read in
    partial_index = match(full_halo_ids, partial_halo_ids)
    halo_was_read = (partial_index>=0)

    # Check that all arrays agree: if we take the full set of halos and discard
    # any which were not read, we should be left with the partial set.
    assert set(partial_data.keys()) == set(full_data.keys())
    for name in partial_data:
        assert np.all(partial_data[name].value == full_data[name][halo_was_read,...])

    # Then we need to check that all halos in the specified part of the sky
    # were read in. All halos which were not read must be outside the radius.
    for i in range(len(halo_was_read)):
        angle = hp.rotator.angdist(vector, full_data["Lightcone/HaloCentre"][i,:])
        assert (angle > radius) or halo_was_read[i]


def try_read_radius_with_soap(vector, radius, remote_dir):
    """
    Read halos in the specified radius about a vector and check that we get
    the right values from SOAP.
    """
    # Read the specified halos
    partial_halos = HaloLightconeFile(halo_lightcone_filename, soap_filename=soap_filename, remote_dir=remote_dir)
    partial_data = partial_halos.read_halos_in_radius(vector, radius, halo_properties+soap_properties)

    # Check SOAP properties against reading the SOAP file directly
    soap_index = partial_data["InputHalos/SOAPIndex"].value
    with partial_halos.open_file(soap_filename) as infile:
        for name in soap_properties:
            assert isinstance(partial_data[name], unyt.unyt_array)
            assert np.all(partial_data[name].value == infile[name][...][soap_index,...])


def test_read_zero_radius(remote_dir):
    """
    Check what happens if we read no halos
    """
    try_read_radius(np.asarray((1, 0, 0), dtype=float), 0.0, halo_properties, remote_dir)


def random_normalized_vectors(N, rng):
    vectors = rng.normal(size=(N, 3))
    norms = np.linalg.norm(vectors, axis=1, keepdims=True)
    normalized_vectors = vectors / norms
    return normalized_vectors


# Generate some randomized test cases
N = 50
rng = np.random.default_rng(seed=0)
test_vectors = random_normalized_vectors(N, rng)
test_radii = rng.random(N) * np.radians(90.0)
test_cases = []
for i in range(N):
    test_cases.append((test_vectors[i,:], test_radii[i]))

@pytest.mark.parametrize("vector,radius", test_cases)
def test_read_radius(vector, radius, remote_dir):
    try_read_radius(vector, radius, halo_properties, remote_dir)

@pytest.mark.parametrize("vector,radius", test_cases)
def test_read_radius_with_soap(vector, radius, remote_dir):
    try_read_radius_with_soap(vector, radius, remote_dir)



#!/bin/env python

import h5py
import healpy as hp
import numpy as np
import unyt
import pytest

from lightcone_io import ParticleLightcone
from lightcone_io.utils import match


# Names of the particle files
particle_filename = "tests/data/particles/lightcone0_particles/lightcone0_0000.{file_nr}.hdf5"
nr_particle_files = 4

# Expected contents of the files
particle_types = ("BH",)
particle_properties = {
    "Coordinates"      : ((0,3), np.float64, 3.08567758e+24),
    "DynamicalMasses"  : ((0,),  np.float32, 1.98841e+43),
    "ExpansionFactors" : ((0,),  np.float32, None),
    "ParticleIDs"      : ((0,),  np.int64,   None),
}

# Range of redshifts covered by the particle lightcone
particle_z_min = 0.0
particle_z_max = 15.0


def test_lightcone_metadata(remote_dir):
    """
    Check that we're reading metadata correctly
    """
    lightcone = ParticleLightcone(particle_filename.format(file_nr=0), remote_dir=remote_dir)
    assert set(lightcone) == set(particle_types)

    # Check we have the expected properties
    for ptype in particle_types:
        for name in lightcone[ptype].properties:
            shape, dtype, units_cgs = particle_properties[name]
            assert lightcone[ptype].properties[name].shape == shape
            assert lightcone[ptype].properties[name].dtype == dtype
            if units_cgs is not None:
                assert np.isclose(units_cgs, (1.0*lightcone[ptype].properties[name].units).in_cgs().value)
            else:
                assert lightcone[ptype].properties[name].units == unyt.dimensionless


def test_read_all(remote_dir):
    """
    Try reading in a full array and comparing to result from h5py
    """
    # Read using the ParticleLightcone class
    lightcone = ParticleLightcone(particle_filename.format(file_nr=0), remote_dir=remote_dir)
    pos1 = lightcone["BH"].read(("Coordinates",))["Coordinates"]
    assert isinstance(pos1, unyt.unyt_array)

    # Read using h5py
    pos2 = []
    for file_nr in range(nr_particle_files):
        with lightcone.open_file(particle_filename.format(file_nr=file_nr)) as infile:
            pos2.append(infile["BH/Coordinates"][...])
    pos2 = np.concatenate(pos2)

    # Compare values
    assert np.all(pos1.value==pos2)


def particle_id(pos):
    """
    Generate unique IDs for the particles in the lightcone, based on positions
    """
    nr_particles = pos.shape[0]
    particle_id = np.ndarray(nr_particles, dtype=np.uint64)
    for i in range(nr_particles):
        particle_id[i] = pos[i,:].astype(np.float64).view(np.uint64).sum()
    # As long as these happen to be unique for the test dataset, we're ok!
    assert len(particle_id) == len(np.unique(particle_id))
    return particle_id


def random_normalized_vectors(N, rng):
    vectors = rng.normal(size=(N, 3))
    norms = np.linalg.norm(vectors, axis=1, keepdims=True)
    normalized_vectors = vectors / norms
    return normalized_vectors

#
# Define some test cases
#
N = 4
rng = np.random.default_rng(seed=0)

# Make an array of direction vectors
vectors = random_normalized_vectors(N, rng)
test_vectors = [None,]
for i in range(N):
    test_vectors.append(vectors[i,:])

# Make an array of angular radii
max_angle = np.radians(90.0)
angles = rng.random(N) * max_angle
test_angles = [None,]
for i in range(N):
    test_angles.append(angles[i])

# Make an array of redshift ranges
test_z_ranges = [None,]
for _ in range(N):
    z1 = rng.random()*particle_z_max
    z2 = rng.random()*particle_z_max
    test_z_ranges.append((min(z1, z2), max(z1, z2)))

@pytest.mark.parametrize("vector,radius", zip(test_vectors, test_angles))
@pytest.mark.parametrize("redshift_range", test_z_ranges)
@pytest.mark.parametrize("exact", (True, False))
def test_read_particles(remote_dir, vector, radius, redshift_range, exact):
    """
    Try reading a redshift range and patch on the sky and comparing to h5py
    """

    # Read the data using the ParticleLightcone class
    lightcone = ParticleLightcone(particle_filename.format(file_nr=0), remote_dir=remote_dir)
    if exact:
        partial_data = lightcone["BH"].read_exact(("Coordinates", "ParticleIDs", "ExpansionFactors"), vector, radius, redshift_range)
    else:
        partial_data = lightcone["BH"].read(("Coordinates", "ParticleIDs", "ExpansionFactors"), vector, radius, redshift_range)

    # Read all of the data using h5py
    full_data = {"Coordinates" : [], "ParticleIDs" : [], "ExpansionFactors" : []}
    for file_nr in range(nr_particle_files):
        with lightcone.open_file(particle_filename.format(file_nr=file_nr)) as infile:
            for name in full_data:
                full_data[name].append(infile["BH"][name][...])
    for name in full_data:
        full_data[name] = np.concatenate(full_data[name])

    # Assign unique particle IDs
    partial_data["UniqueID"] = particle_id(partial_data["Coordinates"])
    full_data["UniqueID"] = particle_id(full_data["Coordinates"])

    # Identify which particles in the full set were read in
    partial_index = match(full_data["UniqueID"], partial_data["UniqueID"])
    particle_was_read = (partial_index>=0)

    # Check that all arrays agree: if we take the full set of particles and discard
    # any which were not read, we should be left with the partial set.
    assert set(partial_data.keys()) == set(full_data.keys())
    for name in partial_data:
        if name != "UniqueID":
            assert np.all(partial_data[name].value == full_data[name][particle_was_read,...])

    # Identify particles in the requested redshift range
    if redshift_range is not None:
        full_z = 1.0/full_data["ExpansionFactors"] - 1.0
        in_z_range = (full_z >= min(redshift_range)) & (full_z <= max(redshift_range))
    else:
        in_z_range = np.ones_like(particle_was_read)

    # Identify particles in the requested patch of sky
    in_radius = np.ones_like(particle_was_read)
    if vector is not None:
        for i in range(len(particle_was_read)):
            angle = hp.rotator.angdist(vector, full_data["Coordinates"][i,:])
            in_radius[i] = (angle <= radius)

    # All particles which were not read should be outside the selection
    selected = in_z_range & in_radius
    assert np.all(particle_was_read | np.logical_not(selected))


@pytest.mark.parametrize("vector,radius", zip(test_vectors, test_angles))
@pytest.mark.parametrize("redshift_range", test_z_ranges)
def test_iterate_chunks(remote_dir, vector, radius, redshift_range):
    """
    Check that the .read() and .iterate_chunks() methods produce the same
    results.
    """
    properties = ("Coordinates", "ParticleIDs", "ExpansionFactors")
    lightcone = ParticleLightcone(particle_filename.format(file_nr=0), remote_dir=remote_dir)

    # Read using .the .read() method
    data1 = lightcone["BH"].read(properties, vector, radius, redshift_range)

    # Read the same data in chunks
    data2 = {name : [] for name in properties}
    for chunk in lightcone["BH"].iterate_chunks(properties, vector, radius, redshift_range):
        assert set(chunk.keys()) == set(properties)
        for name in chunk:
            assert isinstance(chunk[name], unyt.unyt_array)
            data2[name].append(chunk[name])
    for name in data2:
        data2[name] = np.concatenate(data2[name])

    # Compare
    assert set(data1.keys()) == set(data2.keys())
    for name in data1:
        assert np.all(data1[name]==data2[name])




#!/bin/env python

import h5py
import numpy as np
import pytest

from lightcone_io.utils import read_slices, SlicedDatasetReader


# Test cases
nmax = 10000
test_cases = [
    ((0,),         (nmax,)),                   # All in one slice
    ((0, nmax//2), (nmax//2, nmax-nmax//2)),   # All in two slices
    ((),           ()),                        # Zero slices
    ((0,),         (0,)),                      # A zero sized slice
    ((100,),       (200,)),                    # A single partial slice
]


@pytest.mark.parametrize("starts,counts", test_cases)
def test_read_slice(dataset, starts, counts):
    """
    Check that read_slices() reads slices correctly
    """
    # Use read_slices() to read the data
    data1 = read_slices(dataset, starts, counts)

    # Try reading the slices directly with h5py
    data2 = []
    for s, c in zip(starts, counts):
        data2.append(dataset[s:s+c,...])
    if len(data2) > 0:
        data2 = np.concatenate(data2, axis=0)
    else:
        shape = (0,)+dataset.shape[1:]
        data2 = np.zeros(shape, dtype=dataset.dtype)

    # Check the result
    assert data1.shape == data2.shape
    assert data1.dtype == data2.dtype
    assert np.all(data1==data2)


@pytest.mark.parametrize("starts,counts", test_cases)
def test_sliced_dataset_reader(dataset, starts, counts):
    """
    Check that SlicedDatasetReader reads slices correctly
    """
    # Use SlicedDatasetReader to read the data
    data1 = SlicedDatasetReader(starts, counts).read(dataset)

    # Try reading the slices directly with h5py
    data2 = []
    for s, c in zip(starts, counts):
        data2.append(dataset[s:s+c,...])
    if len(data2) > 0:
        data2 = np.concatenate(data2, axis=0)
    else:
        shape = (0,)+dataset.shape[1:]
        data2 = np.zeros(shape, dtype=dataset.dtype)

    # Check the result
    assert data1.shape == data2.shape
    assert data1.dtype == data2.dtype
    assert np.all(data1==data2)



#!/bin/env python

import h5py
import numpy as np
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

particle_filename = "/cosma8/data/dp004/flamingo/Runs/L1000N1800/HYDRO_FIDUCIAL/bh_particle_lightcones/lightcone0_particles/lightcone0_0000.{file_nr}.hdf5"

# Sampling rate to get ~10,000 particles
sample_rate = 7.8125e-07

# Quantities to keep (we discard most to keep the file size down)
names = ("Coordinates", "DynamicalMasses", "ExpansionFactors", "ParticleIDs")

# Where to write the output
output_filename = "/cosma8/data/dp004/jch/test_particles/lightcone0_0000.{file_nr}.hdf5"


def sample_file(file_nr):
    print(f"File {file_nr}")
    rng = np.random.default_rng(seed=file_nr)
    with h5py.File(particle_filename.format(file_nr=file_nr), "r") as infile:
        with h5py.File(output_filename.format(file_nr=file_nr), "w") as outfile:
            infile.copy("InternalCodeUnits", outfile)
            infile.copy("Units", outfile)
            infile.copy("Lightcone", outfile)
            outfile.create_group("BH")
            keep = None
            for name in names:
                print(f"  Dataset: {name}")
                data = infile["BH"][name][...]
                if keep is None:
                    keep = (rng.random(data.shape[0]) < sample_rate)
                data = data[keep,...]
                dset = outfile["BH"].create_dataset(name, data=data) #, compression="gzip", compression_opts=9)
                for attr_name, attr_val in infile["BH"][name].attrs.items():
                    dset.attrs[attr_name] = attr_val

def sample_particles():

    # Determine how many files we have
    with h5py.File(particle_filename.format(file_nr=0), "r") as infile:
        nr_files = infile["Lightcone"].attrs["nr_mpi_ranks"][0]

    # Run in parallel
    with ProcessPoolExecutor(max_workers=8) as executor:
        result = list(executor.map(sample_file, range(nr_files)))


if __name__ == "__main__":
    sample_particles()



#!/bin/env python
#
# This script plots the particles in a pencil beam from a lightcone
#

import matplotlib.pyplot as plt
import matplotlib.colors as col
import numpy as np

import lightcone_io.particle_reader as pr

# Specify one file from the spatially indexed lightcone particle data
input_filename = "/snap8/scratch/dp004/jch/FLAMINGO/ScienceRuns/L2800N5040/HYDRO_FIDUCIAL/indexed_lightcones/lightcone0_particles/lightcone0_0000.0.hdf5"

# Part of the sky to plot
vector = (1.0, 0.0, 0.0)  # Vector pointing at a spot on the sky
radius = np.radians(2.0) # Angular radius around that spot

# Redshift range to plot (set to None for all redshifts in the lightcone)
redshift_range = None

# Open the lightcone
lightcone = pr.IndexedLightcone(input_filename)

plt.figure(figsize=(8,8))

# Loop over particle types
ax = None
nr_types = len(lightcone)
xmax_all = 0
for type_nr, ptype in enumerate(lightcone):
    
    print("Particle type ", ptype)

    # Read in the particle positions and masses
    mass_name = "DynamicalMasses" if ptype=="BH" else "Masses"
    property_names = ("Coordinates", mass_name)
    data = lightcone[ptype].read(property_names=property_names,
                                 redshift_range=redshift_range,
                                 vector=vector, radius=radius,)

    # Unpack positions and masses
    pos  = data["Coordinates"]
    mass = data[mass_name]
    print("Read in %d particles" % pos.shape[0])

    # Find limits for 2D histogram
    xmin = np.amin(pos[:,0])
    xmax = np.amax(pos[:,0])
    ymin = np.amin(pos[:,1])
    ymax = np.amax(pos[:,1])
    aspect=(ymax-ymin)/(xmax-xmin)
    nbins_x = 1000
    nbins_y = int(nbins_x*aspect) # for square-ish pixels
    xmax_all = max(xmax, xmax_all)

    # Make a plot for this particle type
    ax = plt.subplot(nr_types, 1, type_nr+1, sharex=ax, sharey=ax)
    plt.hist2d(pos[:,0], pos[:,1], bins=(nbins_x, nbins_y), weights=mass, norm=col.LogNorm())
    plt.gca().set_aspect("equal")
    plt.xlabel("x [Mpc]")
    plt.ylabel("y [Mpc]")
    plt.title(ptype)

    del pos
    del mass
    del data

plt.xlim(0, xmax_all)
plt.tight_layout()
plt.savefig("pencil_beam.png")
plt.show()



#!/bin/env python

import re
import io
import os.path

import numpy as np

import h5py
import h5py.h5t
import h5py.h5s


def create_dataset(loc, name, type_id, shape, dcpl_id):
    """
    Create a new HDF5 dataset using supplied type and creation properties
    """    
    
    space_id = h5py.h5s.create_simple(shape, (h5py.h5s.UNLIMITED,)*len(shape))
    dset_id = h5py.h5d.create(loc.id, name.encode(), type_id, space_id, dcpl_id)
    space_id.close()
    dset_id.close()


def gather_lists(comm, l):
    l = comm.allgather(l)
    return [item for sublist in l for item in sublist]


class LightconeMetadata:

    def index_file_name(self, basedir=None):
        if basedir is None:
            basedir = self.basedir
        return "%(basedir)s/%(basename)s_index.hdf5" % {"basedir" : basedir, "basename" : self.basename}

    def particle_file_name(self, mpi_rank_nr, file_nr, basedir=None):
        if basedir is None:
            basedir = self.basedir
        return ("%(basedir)s/%(basename)s_particles/%(basename)s_%(file_nr)04d.%(mpi_rank_nr)d.hdf5" %
                {"basedir" : basedir, "basename" : self.basename,
                 "file_nr" : file_nr, "mpi_rank_nr" : mpi_rank_nr})

    def __init__(self, basedir, basename, comm=None):

        self.basedir = basedir
        self.basename = basename
        self.file = None

        if comm is None:
            comm_size = 1
            comm_rank = 0
        else:
            comm_rank = comm.Get_rank()
            comm_size = comm.Get_size()

        # Read the index to determine how many files there are
        if comm_rank == 0:
            with h5py.File(self.index_file_name(), "r") as infile:
                self.nr_mpi_ranks = infile["Lightcone"].attrs["nr_mpi_ranks"][0]
                self.final_particle_file_on_rank = infile["Lightcone"].attrs["final_particle_file_on_rank"]
                self.part_types = []
                for name in infile["Lightcone"].attrs:
                    m = re.match(r"minimum_redshift_(.*)", name)
                    if m:
                        self.part_types.append(m.group(1))
        else:
            self.nr_mpi_ranks = None
            self.final_particle_file_on_rank = None
            self.part_types = None
        if comm is not None:
            self.nr_mpi_ranks = comm.bcast(self.nr_mpi_ranks)
            self.final_particle_file_on_rank = comm.bcast(self.final_particle_file_on_rank)
            self.part_types = comm.bcast(self.part_types)

        # Generate the full list of filenames to read on this rank
        self.filenames = []
        count = 0
        for mpi_rank_nr in range(self.nr_mpi_ranks):
            for file_nr in range(self.final_particle_file_on_rank[mpi_rank_nr]+1):
                if count % comm_size == comm_rank:
                    filename = self.particle_file_name(mpi_rank_nr, file_nr)                
                    self.filenames.append(filename)
                count += 1
        self.nr_files = len(self.filenames)

        # Make an in-memory HDF5 file
        if comm_rank == 0:
            bio = io.BytesIO()
            tmpfile = h5py.File(bio, 'w')
            for ptype in self.part_types:
                tmpfile.create_group(ptype)

        # Determine number of particles of each type in each file
        self.nr_particles_file = {name : [] for name in self.part_types}
        for filename in self.filenames:
            with h5py.File(filename, "r") as infile:
                for ptype in self.part_types:
                    if ptype in infile:
                        nr = infile[ptype]["ExpansionFactors"].shape[0]
                    else:
                        nr = 0
                    self.nr_particles_file[ptype].append(nr)

        # In MPI mode we now need to combine the results from different ranks
        if comm is not None:            
            self.filenames = gather_lists(comm, self.filenames)
            for ptype in self.part_types:
                self.nr_particles_file[ptype] = gather_lists(comm, self.nr_particles_file[ptype])
            self.nr_files = len(self.filenames)

        # Calculate total number of particles of each type
        self.nr_particles_total = {name : 0 for name in self.part_types}
        for ptype in self.part_types:
            self.nr_particles_total[ptype] = sum(self.nr_particles_file[ptype])

        # Determine which quantities exist for each particle type by
        # opening a file which contains particles of each type.
        if comm_rank == 0:
            self.properties = {}
            # Loop over particle types
            for ptype in self.part_types:
                # Loop over files to find one that contains particles of this type
                for i in range(self.nr_files):
                    if self.nr_particles_file[ptype][i] > 0:
                        # This file contains particles of this type
                        self.properties[ptype] = []
                        infile = h5py.File(self.filenames[i], "r")
                        for name in infile[ptype]:
                            dset = infile[ptype][name]
                            if "a-scale exponent" in dset.attrs:
                                # Create a zero sized in-memory dataset with the same data type
                                # and dataset creation property list. This preserves the filters,
                                # custom data types and and compression settings from the original
                                # dataset. Also copy any attributes.
                                dtype_id = dset.id.get_type()
                                dcpl_id = dset.id.get_create_plist()
                                shape = dset.shape[1:]
                                create_dataset(tmpfile[ptype], name, dtype_id, (0,)+shape, dcpl_id)
                                for attr_name in dset.attrs:
                                    tmpfile[ptype][name].attrs[attr_name] = dset.attrs[attr_name]
                                self.properties[ptype].append(name)
                        infile.close()
                        break
        else:
            self.properties = None
        if comm is not None:
            self.properties = comm.bcast(self.properties)

        # Store image of the in-memory file: this puts the information about special data types,
        # filters etc into a form that we can MPI_Bcast().
        if comm_rank == 0:
            tmpfile.close()
            self.file_image = bio.getvalue()
        else:
            self.file_image = None
        if comm is not None:
             self.file_image = comm.bcast(self.file_image)

        # Find offset to first particle in each file
        self.offset_file = {}
        for ptype in self.part_types:
            self.nr_particles_file[ptype] = np.asarray(self.nr_particles_file[ptype], dtype=np.int64)
            self.offset_file[ptype] = np.cumsum(self.nr_particles_file[ptype]) - self.nr_particles_file[ptype]

    def open_image(self):
        self.file = h5py.File(io.BytesIO(self.file_image), "r")

    def __del__(self):
        if self.file is not None:
            #self.file.close()
            self.file = None
